#!/usr/env/bin python

import argparse
import enum
import functools
import importlib
import inspect
import logging
import multiprocessing
import json
import os
import signal
import time
import typing
import types

from datetime import datetime

from bert import utils as bert_utils, constants as bert_constants
from bert.deploy import utils as bert_deploy_utils

logger = logging.getLogger(__name__)

STOP_DAEMON: bool = False
class Service(enum.Enum):
    AWSLambda: str = 'aws-lambda'

def capture_options() -> typing.Any:
    parser = argparse.ArgumentParser()
    parser.add_argument('-s', '--service', default=Service.AWSLambda, type=Service)
    parser.add_argument('-u', '--undeploy', action="store_true", default=False)
    parser.add_argument('-i', '--invoke', action="store_true", default=False)
    parser.add_argument('-a', '--invoke-async', default=False, action='store_true')
    parser.add_argument('-m', '--module-name', default='bert')
    parser.add_argument('-f', '--flush', default=False, action="store_true")
    parser.add_argument('-d', '--dry-run', default=False, action="store_true", help="Create the lambda functions and output ./lambdas without deploying to AWS")
    parser.add_argument('-t', '--run-monitor', default=False, action="store_true", help="Run Monitor function locally")
    return parser.parse_args()

def deploy_service(options) -> None:
    jobs: typing.Dict[str, typing.Any] = bert_utils.scan_jobs(options.module_name)
    jobs: typing.Dict[str, typing.Any] = bert_utils.map_jobs(jobs, options.module_name)

    if options.service == Service.AWSLambda:
        if options.invoke:
            import boto3
            job_name: str = [job for job in jobs.keys()][0]
            invoke_args: typing.List[typing.Dict[str, typing.Any]] = {key: value for key, value in jobs.items()}[job_name]['aws-deploy']['invoke-args']
            client = boto3.client('lambda')
            if len(invoke_args) < 1:
                logger.info(f'Invoking Job[{job_name}]')
                client.invoke(FunctionName=job_name, InvocationType='Event')

            else:
                logger.info(f'Invoking Job[{job_name}] with {len(invoke_args)} payloads.')
                if options.invoke_async:
                    for args in invoke_args:
                        payload: bytes = json.dumps({'bert-inputs': [args]}).encode(bert_constants.ENCODING)
                        client.invoke(FunctionName=job_name, InvocationType='Event', Payload=payload)
                else:
                    payload: bytes = json.dumps({'bert-inputs': invoke_args}).encode(bert_constants.ENCODING)
                    client.invoke(FunctionName=job_name, InvocationType='Event', Payload=payload)

            import sys; sys.exit(0)

        if options.run_monitor:
            from bert.deploy import reporting
            logger.info(f'Running monitor function locally')
            reporting.monitor_function_progress(options.module_name)
            import sys; sys.exit(0)

        bert_deploy_utils.validate_inputs(jobs)
        bert_deploy_utils.build_project(jobs)
        bert_deploy_utils.build_lambda_handlers(jobs)
        bert_deploy_utils.build_archives(jobs)
        bert_deploy_utils.create_roles(jobs)
        bert_deploy_utils.scan_dynamodb_tables(jobs)
        bert_deploy_utils.destroy_lambda_to_table_bindings(jobs)
        bert_deploy_utils.destroy_lambda_concurrency(jobs)
        bert_deploy_utils.destroy_sns_topic_lambdas(jobs)
        bert_deploy_utils.destroy_lambdas(jobs)
        if options.flush:
            bert_deploy_utils.destroy_dynamodb_tables(jobs)

        bert_deploy_utils.create_lambdas(jobs)
        bert_deploy_utils.create_lambda_concurrency(jobs)
        bert_deploy_utils.create_dynamodb_tables(jobs)
        bert_deploy_utils.create_reporting_dynamodb_table()
        bert_deploy_utils.bind_lambdas_to_tables(jobs)
        bert_deploy_utils.bind_events_for_bottle_functions(jobs)
        bert_deploy_utils.bind_events_for_init_function(jobs)
        bert_deploy_utils.destroy_monitor()
        bert_deploy_utils.deploy_monitor(options.module_name)

    else:
        raise NotImplementedError(options.service)

def run_from_cli():
    import sys, os
    sys.path.append(os.getcwd())
    options = capture_options()
    deploy_service(options)

if __name__ in ['__main__']:
    run_from_cli()




import base64
import boto3
import json
import logging
import os
import typing
import uuid

from bert import \
    exceptions as bert_exceptions, \
    constants as bert_constants

from botocore.exceptions import ClientError

logger = logging.getLogger(__name__)
PWN: typing.TypeVar = typing.TypeVar('PWN')

class assume_role:
    __slots__ = ('_role_arn', '_duration', '_session_name', '_env_vars', '_old_values')
    _env_vars: typing.Dict[str, str]
    _old_values: typing.Dict[str, str]
    _role_arn: str
    _duration: int
    _session_name: str

    def __init__(self: PWN, role_arn: str, duration: int = 3600) -> None:
        self._role_arn = role_arn
        self._duration = duration
        self._session_name = str(uuid.uuid4())
        self._env_vars = {}
        self._old_values = {}

    def __enter__(self: PWN) -> PWN:
        sts_client = boto3.client('sts')
        try:
            response = sts_client.assume_role(RoleArn=self._role_arn, RoleSessionName=self._session_name, DurationSeconds=self._duration)
        except ClientError as err:
            # This could happen for two reasons.
            # 1. The user is not registered as a trust entity
            # 2. The duration of the assume role is greater than what has been configured in IAM
            raise err

        self._env_vars['AWS_ACCESS_KEY_ID'] = response['Credentials']['AccessKeyId']
        self._env_vars['AWS_SECRET_ACCESS_KEY'] = response['Credentials']['SecretAccessKey']
        self._env_vars['AWS_SESSION_TOKEN'] = response['Credentials']['SessionToken']
        for key, value in self._env_vars.items():
            self._old_values[key] = os.environ.get(key, None)
            os.environ[key] = value

        return self

    def __exit__(self: PWN, exception_type: 'ExceptionType', exception_value: Exception, traceback: typing.Any) -> None:
        for key, old_value in self._old_values.items():
            if old_value is None:
                del os.environ[key]

            else:
                os.environ[key] = old_value

        if exception_value:
            raise exception_value

class kms:
    _kms_client: 'boto3.client("kms")'
    _alias: str
    _key: typing.Dict[str, typing.Any]
    _auto_create: bool
    _usernames: typing.List[str]
    _header: str = 'bert-etl-encrypted'
    class CipherEncryptionError(bert_exceptions.BertException):
        pass
    class CipherDecryptionError(bert_exceptions.BertException):
        pass
    class KeyNotFound(bert_exceptions.BertException):
        pass

    def __init__(self: PWN, alias: str, usernames: typing.List[str] = [], auto_create: bool = False) -> None:
        self._kms_client = boto3.client('kms')
        self._alias = alias
        self._key = None
        self._auto_create = auto_create
        self._usernames = usernames

    def _gen_key_policy(self: PWN) -> str:
        caller_info: typing.Dict[str, typing.Any] = boto3.client('sts').get_caller_identity()
        caller_username: str = caller_info['Arn'].rsplit(':', 1)[-1]
        usernames: typing.List[str] = list({username for username in self._usernames})
        if not caller_username in usernames:
            logger.info(f'Adding AWS Caller username[{caller_username}] to KMSKey[{self._alias}]')
            usernames.append(caller_username)

        if not 'root' in usernames:
            logger.info(f'Adding AWS Caller username[root] to KMSKey[{self._alias}] Usernames[{",".join(usernames)}]')
            usernames.append('root')

        return json.dumps({
            'Version': '2012-10-17',
            'Id' : 'key-default-1',
            'Statement' : [
                {
                    'Sid' : 'Enable IAM User Permissions',
                    'Effect' : 'Allow',
                    'Principal' : {
                        'AWS' : f'arn:aws:iam::{caller_info["Account"]}:{username}'
                    },
                    'Action' : 'kms:*',
                    'Resource' : '*'
                } for username in usernames]
        })

    def _load_key_by_alias(self: PWN, key_alias: str) -> typing.Dict[str, typing.Any]:
        for page in self._kms_client.get_paginator('list_aliases').paginate():
            for kms_alias in page['Aliases']:
                if kms_alias['AliasName'] == key_alias:
                    return self._kms_client.describe_key(KeyId=kms_alias['TargetKeyId'])

    def __enter__(self: PWN) -> PWN:
        alias_name: str = f'alias/{self._alias}'
        if self._key is None:
            self._key = self._load_key_by_alias(alias_name)

        if self._auto_create and self._key is None:
            key = self._kms_client.create_key(
                Description=f'{self._alias} Key',
                Origin="AWS_KMS",
                KeyUsage='ENCRYPT_DECRYPT')
            key_policy: str = self._gen_key_policy()
            self._kms_client.create_alias(
                AliasName=f'alias/{self._alias}',
                TargetKeyId=key['KeyMetadata']['KeyId'])
            self._kms_client.put_key_policy(
                KeyId=key['KeyMetadata']['KeyId'],
                PolicyName='default',
                Policy=key_policy)

            self._key = self._load_key_by_alias(alias_name)

        if self._key is None:
            raise self.__class__.KeyNotFound(self._alias)

        return self

    def __exit__(self: PWN, exception_type: 'ExceptionType', exception_value: Exception, traceback: typing.Any) -> None:
        if exception_value:
            raise exception_value

    def update_usernames(self: PWN) -> None:
        key_policy: str = self._gen_key_policy()
        self._kms_client.put_key_policy(
            KeyId=self._key['KeyMetadata']['KeyId'],
            PolicyName='default',
            Policy=key_policy)

    def decrypt(self: PWN, value: str) -> str:
        if not value.startswith(self._header):
            raise self.__class__.CipherDecryptionError('Value already decrypted')

        header, alias, cipher = value.split(':')
        encrypted_value: bytes = base64.b64decode(cipher.encode(bert_constants.ENCODING))
        try:
            decrypted_value: typing.Dict[str, typing.Any] = self._kms_client.decrypt(CiphertextBlob=encrypted_value)
        except self._kms_client.exceptions.InvalidCiphertextException:
            raise NotImplementedError

        return decrypted_value['Plaintext'].decode(bert_constants.ENCODING)

    def encrypt(self: PWN, value: str) -> str:
        if value.startswith(self._header):
            raise self.__class__.CipherEncryptionError('Value already encrypted')

        encrypted_value: typing.Dict[str, typing.Any] = self._kms_client.encrypt(
            KeyId=self._key['KeyMetadata']['KeyId'],
            Plaintext=value.encode(bert_constants.ENCODING))
        ascii_value: str = base64.b64encode(encrypted_value['CiphertextBlob']).decode(bert_constants.ENCODING)
        return f'{self._header}:{self._alias}:{ascii_value}'




import boto3
import logging
import typing
import uuid

logger = logging.getLogger(__name__)
PWN: typing.TypeVar = typing.TypeVar('PWN')

def aws_account_id() -> str:
    sts_client = boto3.client('sts')
    return sts.get_caller_identity()['Account']

def map_iam_role_by_arn(role_arn: str) -> typing.Dict[str, typing.Any]:
    if role_arn is None:
        return {'Arn': None}

    iam_client = boto3.client('iam')
    for page in iam_client.get_paginator('list_roles').paginate(PathPrefix='/'):
        for role in page['Roles']:
            if role_arn == role['Arn']:
                return role

    # Match the interface expected
    return {'Arn': None}

def map_iam_role(role_name: str) -> typing.Dict[str, typing.Any]:
    iam_client = boto3.client('iam')
    for page in iam_client.get_paginator('list_roles').paginate(PathPrefix='/'):
        for role in page['Roles']:
            if role['RoleName'] == role_name:
                return role

def map_iam_policy(policy_name: str) -> typing.Dict[str, typing.Any]:
    iam_client = boto3.client('iam')
    for page in iam_client.get_paginator('list_policies').paginate(PathPrefix='/'):
        for policy in page['Policies']:
            if policy['PolicyName'] == policy_name:
                return policy

